/* readseqXCMD.c 
 *
 *
 * Copyright 1990 by d.g.gilbert.
 * dogStar Software && Indiana University Biology Dept.
 * email: gilbertd@iubio.bio.indiana.edu
 *
 * from readseq.c
 *
 * Language: MPW-C 3.0
 * 
	
	 HCard syntax is:
--
--	readseq [#], 		: if 0, return list of seqs, else fetch seq #
--		[iscardfld, name fld], : fld for filename (in/out)
--		[iscardfld, doc fld],	: fld for seq doc (out)
--		[file data]						: container w/ all of file data (in)

*/

/*-------- Makefile

C -b readseqXCMD.c -mbg off
Link -w -rt XCMD=10010 
	-m ENTRYPOINT 
	-sg readseq 
	readseqXCMD.c.o 
	"{Libraries}HyperXLib.o" 
	"{Libraries}Interface.o" 
	"{CLibraries}StdCLib.o" 
	"{Clibraries}"Cinterface.o 
	"{CLibraries}CRuntime.o" 
	-o "HyperGCG" 
#	-o "readseq.xcmd" -x readseq.xref	-map > readseq.map
#-------------------------*/



#include <CType.h>
#include <String.h>
#include <Types.h>
#include <StdLib.h>
/* #include <StdIO.h> */
#include <HyperXCmd.h>

#include <Memory.h>
#include <OSUtils.h>
#include <Files.h>
#include <ToolUtils.h>
#include <Packages.h>



/* XCMD entry -- MUST BE FIRST CODE IN SEG */

void 	CallReadSeq(XCmdPtr pXCmd);

pascal void EntryPoint(XCmdPtr pXCmd)
{
	CallReadSeq(pXCmd);
	return;
}


typedef	char	boolean;
#define NEWLINE       '\n'  
#define	false	0
#define	true	1
#define min(a,b)  		(a<b)?a:b
#define max(a,b)  		(a>b)?a:b
#define _toupper(c)		((c)-'a'+'A')
#define _tolower(c)		((c)-'A'+'a')
#define	Isalpha(c)		((c >= 'A' & c <= 'Z') \
										|  (c >= 'a' & c <= 'z'))


#define MAXSEQ			32000		/* HCard field max chars */
#define MAXDOC			10000		/* this is enough to read... */

/* readSeq errors */
#define eFileNotFound  	-1
#define	eNoData 				-2
#define	eMemFull 				-3
#define eItemNotFound  	-4
#define eWrongFormat		-5

/* formats known to readSeq */
#define kIG             1
#define kGenBank        2
#define kNBRF           3
#define kEMBL           4
#define kGCG            5
#define kStrider        6
#define kFitch          7
#define kPearson        8
#define kZuker          9
#define kOlsen          10
#define kUnknown        11
#define kMinFormat      1
#define kMaxFormat      11



int Toupper(char c)
{
	return( (c>='a')&&(c<='z') ? (c-('a'-'A')) : c );
}


void upcase(char *s)
{
	int i,l;
	l = strlen(s);
	for (i=0;i<l;i++) s[i] = Toupper(s[i]);
}



#define Local      static    /* local functions */

#define stdsymbols	"_.-*?" 
#define allsymbols  "_.-*?<>{}[]()!@#$%^&=+;:'/|`~\"\\" 

#define isseqchar(c) (Isalpha(c) || strchr(allsymbols, c))
/* 
#define isseqchar(c) (Isalpha(c) || c == '_' || c == '.' \
											|| c == '-'|| c == '*' || c == '?')
*/


typedef	struct {  /* replaces stdIO FILE */
	char	*d;  	/* ptr to data */
	long	i;	 	/* index into d */
	long	len;	/* length of d ? */
	} File;

/* Local variables for readSeq: */
struct LOC_readSeq {
		/* parameters */
  short choice, *nseq, *format;
  long *seqlen, *doclen;
  char *seq, *seqid, *doc;
		/* locals */
  boolean allDone, done, addit, doadd;
  long 	maxseq, ninfo, maxdoc;
	short	err;
	File 	f;  
  char s[256], *cp;
};


/*---
#define getline(V)	 { \
			fgets(V->s, 256, V->f); \
      V->cp = strchr(V->s, '\n'); \
      if (V->cp != NULL) *V->cp = 0; \
			}
----*/

#define Feof(f)			(f.d[f.i] == 0 | f.i >= f.len)
#define Frewind(f)	{f.i = 0;}
#define Fclose(f)		{f.i = 0;}

Local void getline(struct LOC_readSeq *V)
{
	long	i, si;
	
	i = V->f.i;
	si = 0;
	while (si<256 & V->f.d[i] != '\n' & V->f.d[i] != 0) 
		V->s[si++] = V->f.d[i++];
	V->s[si] = 0;
	if (V->f.d[i] == '\n') i++;
	V->f.i = i;  
}


Local void addseq(char *s, struct LOC_readSeq *V)
{
	while (*s != 0) {	 
		if (isseqchar(*s) && V->addit) {
			if (*V->seqlen >= V->maxseq) V->err = eMemFull;
			else {
				V->seq[(*V->seqlen)++] = *s;
				/* ?! want spacers for HyperGCG */
				if (*V->seqlen % 55 == 54)
					V->seq[(*V->seqlen)++] = '\n';
				else if (*V->seqlen % 11 == 10)
					V->seq[(*V->seqlen)++] = ' ';
				}
		  }
		s++;
		}
}  /*addseq*/



Local void addinfo(char *s, struct LOC_readSeq *V)
{
  char s2[256], *si;

	si = s2;
	while (*s == ' ') s++; 
  /* sprintf(si, " %d)  %s", *V->nseq, s); */
	numtostring(*V->nseq, s2);
	strcat(s2, ")  ");
	strcat(s2, s);
	while ((*si != 0) && (V->ninfo < V->maxseq)) 	 
  	{ V->seq[V->ninfo++] = *si; 
		  si++; } 
	if (V->ninfo < V->maxseq) V->seq[V->ninfo++] = NEWLINE;
}  /*addinfo*/


Local void adddoc(char *s, struct LOC_readSeq *V)
{
	while ((*s != 0) && (*V->doclen < V->maxdoc)) 	 
  	{ V->doc[(*V->doclen)++] = *s++; } 
	if (*V->doclen < V->maxdoc) V->doc[(*V->doclen)++] = NEWLINE;
}  /*adddoc*/



Local void readLoop(short margin, boolean addfirst,
		    		boolean (*endTest)(char *s, boolean *addend), 
						struct LOC_readSeq *V)
{
	boolean addend = false;
	
  (*V->nseq)++;
  *V->seqlen = 0;
  if (V->choice == 0) V->doadd = false;
  else V->doadd = (*V->nseq == V->choice);

  if (addfirst) addseq(V->s, V);  
  do {
    getline(V);	
    V->done = Feof(V->f);				
    V->done |= (*endTest)(V->s, &addend);
    if (V->doadd && (addend || !V->done)) {
      addseq( (V->s)+margin, V);
    }
  } while (!V->done);
	
  if (V->choice == 0) addinfo(V->seqid, V);
  else V->allDone = (*V->nseq >= V->choice);
}  /*readLoop*/



Local boolean endIG(char *s, boolean *addend)
{ 	
	*addend = true; /* 1 or 2 occur in line w/ bases */
	return((strchr(s,'1')!=NULL) || (strchr(s,'2')!=NULL)); 
}

Local void readIG(struct LOC_readSeq *V)
{
	
  /*IG -- many seqs/file */
  *V->format = kIG;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
		do { 
			getline(V);
			if (*V->s == ';') adddoc((V->s)+1, V);
		} while (! (Feof(V->f) || ((*V->s != 0) && (*V->s != ';')) ));
    if (Feof(V->f))
      V->allDone = true;
    else {
  		strcpy(V->seqid, V->s);
			readLoop(0, false, endIG, V);
    }
  }
}  /*readIG*/



Local boolean endStrider(char *s, boolean *addend)
{ 
	*addend = false;
	return (strstr( s, "//") != NULL);  
} 

Local void readStrider(struct LOC_readSeq *V)
{ /* ? only 1 seq/file ? */

  *V->format = kStrider;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
		getline(V);
    strcpy(V->seqid, (V->s)+1);
		adddoc((V->s)+1, V);
		while ((!Feof(V->f)) && (*V->s == ';')) {  
      getline(V);
			if (*V->s == ';') adddoc((V->s)+1, V);
			}
    if (Feof(V->f)) V->allDone = true;
		else readLoop(0, true, endStrider, V);
  }
}  /*readStrider*/


Local boolean endGB(char *s, boolean *addend)
{ 
	*addend = false;
	return ((strstr(s,"//") != NULL) || (strstr(s,"LOCUS") == s));
}

Local void readGenBank(struct LOC_readSeq *V)
{ /*GenBank -- many seqs/file */

  *V->format = kGenBank;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    strcpy(V->seqid, (V->s)+12);
		adddoc((V->s), V);
		while (! (Feof(V->f) || strstr(V->s,"ORIGIN") == V->s)) {
		 	getline(V);
			adddoc((V->s), V);
			}
    readLoop(0, false, endGB, V);
    if (!V->allDone) {
     while (! (Feof(V->f) || ((*V->s != 0) 
			 && (strstr( V->s,"LOCUS") == V->s)))) 
				getline(V);
			}
    if (Feof(V->f)) V->allDone = true;
  }
}  /*readGenBank*/


Local boolean endNBRF(char *s, boolean *addend)
{		
	char	*a;
	
	if ((a = strchr(s, '*')) != NULL) { /* end of 1st seq */
		/* "*" can be valid base symbol, drop it here */
		*a = 0;
		*addend = true;
		return(true);
		}
	else if (*s == '>') {	/* start of next seq */
		*addend = false;
		return(true);
		}
	else
		return(false);
}

Local void readNBRF(struct LOC_readSeq *V)
{

  *V->format = kNBRF;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    strcpy(V->seqid, (V->s)+4);
		adddoc((V->s), V);
    getline(V);   /*skip title-junk line*/
		adddoc((V->s), V);
    readLoop(0, false, endNBRF, V);
    if (!V->allDone) {
     while (!(Feof(V->f) || (*V->s != 0 && *V->s == '>'))) 
				getline(V);
    	}
    if (Feof(V->f)) V->allDone = true;
  }
}  /*readNBRF*/



Local boolean endPearson(char *s, boolean *addend)
{		
	*addend = false;
	return(*s == '>');
}

Local void readPearson(struct LOC_readSeq *V)
{

  *V->format = kPearson;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    strcpy(V->seqid, (V->s)+1);
		adddoc((V->s)+1, V);
    readLoop(0, false, endPearson, V);
    if (!V->allDone) {
     while (!(Feof(V->f) || ((*V->s != 0) && (*V->s == '>')))) 
			 	getline(V);
    	}
    if (Feof(V->f)) V->allDone = true;
  }
}  /*readPearson*/



Local boolean endEMBL(char *s, boolean *addend)
{	
	*addend = false;
	return( strstr(s,"ID   ") == s);
}

Local void readEMBL(struct LOC_readSeq *V)
{

  *V->format = kEMBL;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    strcpy(V->seqid, (V->s)+5);
 		adddoc((V->s), V);
    do {
      getline(V); 
 			adddoc((V->s), V);
    } while (!(Feof(V->f) | (strstr(V->s,"SQ   ") == V->s)));

    readLoop(0, false, endEMBL, V);
    if (!V->allDone) {
      while (!(Feof(V->f) |
	       ((*V->s != '\0') & (strstr(V->s,"ID   ") == V->s)))) 
			getline(V);
    }
    if (Feof(V->f)) V->allDone = true;
  }
}  /*readEMBL*/



Local boolean endZuker(char *s, boolean *addend)
{	
	*addend = false;
	return( *s == '(' );
}

Local void readZuker(struct LOC_readSeq *V)
{
  /*Zuker -- many seq/file ?*/
  /*! 1st string is Zuker's Fortran format */

  *V->format = kZuker;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    getline(V);  /*s == "seqLen seqid string..."*/
    strcpy(V->seqid, (V->s)+6);
 		adddoc((V->s)+6, V);
    readLoop(0, false, endZuker, V);
    if (!V->allDone) {
      while (!(Feof(V->f) | 
				((*V->s != '\0') & (*V->s == '('))))  
					getline(V);
			}
    if (Feof(V->f)) V->allDone = true;
  }
}  /*readZuker */



Local boolean endFitch(char *s, boolean *addend)
{	
	/* this is a somewhat shaky end, 
		1st char of line is non-blank for seq. title 
	*/
	*addend = false;
	return( *s != ' ' ); 
}

Local void readFitch(struct LOC_readSeq *V)
{
  boolean first;

  *V->format = kFitch;
  first = true;
  while (!V->allDone) {
		*V->doclen = 0; /* zero for each seq */
    /*Fitch -- many seqs/file */
    if (!first) strcpy(V->seqid, V->s);
    readLoop(0, first, endFitch, V);
    if (Feof(V->f)) V->allDone = true;
    first = false;
  }
 	adddoc((V->seqid), V);
}  /*readFitch */


Local void readUnknown(struct LOC_readSeq *V)
{  /*Unknown -- 1seq/file*/

  *V->format = kUnknown;   
  (*V->nseq)++;
  *V->seqlen = 0;
  V->addit = (V->choice > 0);
  addseq(V->seqid, V);   /*from above..*/
  strcpy(V->seqid," [Unknown form]");
  do {
    addseq(V->s, V);
    V->done = Feof(V->f);
    getline(V);
  } while (!V->done);
  if (V->choice == 0) addinfo(V->seqid, V);
  V->allDone = true;
}  /*readUnknown */


Local void readUWGCG(struct LOC_readSeq *V)
{  /* UWGCG -- 1seq/file*/
	char	*si;
	
  *V->format = kGCG;
  (*V->nseq)++;
  *V->seqlen = 0;
  V->addit = (V->choice > 0);
	strcpy(V->seqid, V->s);
	/*drop ".." from id*/
	if (si = strstr(V->seqid,"..")) *si = 0; 
 	adddoc((V->seqid), V);  
  do {
    V->done = Feof(V->f);
    getline(V); 
    if (!V->done) addseq((V->s), V);
  } while (!V->done);
  if (V->choice == 0) addinfo(V->seqid, V);
  V->allDone = true;
}  /*readUWGCG */


Local void readOlsen(struct LOC_readSeq *V)
{ /* G. Olsen /print output from multiple sequence editor */
	
	char		*si, *sj, *sk, sid[40];
	boolean	indata = false;
	
	*V->format = kOlsen;
	*V->seqlen = 0;
  V->addit = (V->choice > 0);
	Frewind(V->f);
	do {
		getline(V);
		V->done = Feof(V->f);

		if (indata) {
			if (strstr(V->s, sid)) {	
				/* ! always left margin of 22 */
				/* variable right id margin, drop id-2 spaces at end */
				/* Spaces are valid alignment data !! */
				si = (V->s)+22; 			
				if (sk = strstr(si, sid)) *(sk-2) = 0;  
				for (sk = si; *sk != 0; sk++) if (*sk == ' ') *sk = '.';
				addseq(si, V);
				}
			}
		else if (sk = strstr(V->s, "): ")) {  /* seq info header line */
  		(*V->nseq)++;
			si = 1 + strchr(V->s,'(');
			*sk = ' ';
			if (V->choice == 0) addinfo( si, V);
			else if (*V->nseq == V->choice) {
				strcpy(V->seqid, si);
				sj = strchr(V->seqid, ':');
				while (*(--sj) == ' ') ;
				while (--sj != V->seqid) { if (*sj == ' ') *sj = '_'; }
					
				*sk = 0;
				while (*(--sk) == ' ') *sk = 0;
				strcpy(sid, si); 
				}
			}
		else if (strstr(V->s,"identity:   Data:")) { 
			indata = true;
			if (V->choice == 0) V->done = true;
			}
			
  } while (!V->done);
			
  V->allDone = true;
} /*readOlsen*/




short readSeq(
			short choice_, 	/*0=list seqs in file, >0 = read item*/
			char 	*fdata, 	/*!! all file data*/
			long	flen,			/* length of fdata */
			char  *seq_, 		/*sequence or list storage*/
			long  *seqlen_, /*in: max seq size, out: seq size*/
	    short *nseq_,   /*out (list): # seqs in file*/
			short *format_, /*out (list), in (>0): seq format*/
			char	*doc_,		/*out (seq): doc from seq*/
			long	*doclen_,	/*in: max doc size, out: doc size */
			char  *seqid_)  /*out (seq): seq info*/
  										/*return error: 0=okay*/
{
  struct LOC_readSeq V;
  boolean gotuw, gotolsen, isfitch;
  short i, l;

	
  V.choice = choice_;
  V.seq = seq_;
  V.seqlen = seqlen_;
  V.nseq = nseq_;
  V.format = format_;
  V.seqid = seqid_;
  V.maxseq = *V.seqlen-1; /* leave room for '\0' */
  *V.seqlen = 0;
  V.ninfo = 0;
  *V.nseq = 0;
  *V.seqid = '\0';
	V.doc	= doc_;
	V.doclen = doclen_;
	V.maxdoc = *V.doclen-1;
	*V.doclen = 0;
  V.err = 0;
  if (V.choice < 0) V.choice = 0;
  if (V.choice == 0 || *V.format < kMinFormat || *V.format > kMaxFormat)
    *V.format = 0;
  V.addit = (V.choice > 0);
  V.allDone = false;


  /*  -- file open, replace w/ handle file --
	strcpy(V.fname, fname_);  
  V.f = NULL;  
  V.f = fopen(V.fname, "r");
  if (V.f == NULL) V.err = eFileNotFound;
	--------------- */
	V.f.d = fdata;
	V.f.i = 0;
	V.f.len = flen; /* ?? */
  if (V.f.len == 0) V.err = eFileNotFound;

	if (V.err != 0) ;
  else if (*V.format > 0) { /*don't re-check format*/
    do {
      getline(&V);
      l = strlen(V.s);
      while (l > 0 && V.s[--l] == ' ') ;
    } while (!((l > 0) | Feof(V.f)));
		
    if (l == 0) V.err = eNoData;
    else switch (*V.format) {

      case kUnknown:
				readUnknown(&V);
				break;

      case kIG:
				readIG(&V);
				break;

      case kStrider:
				readStrider(&V);
				break;

      case kGenBank:
				readGenBank(&V);
				break;

      case kNBRF:
				readNBRF(&V);
				break;

      case kPearson:
				readPearson(&V);
				break;

      case kEMBL:
				readEMBL(&V);
				break;

      case kZuker:
				readZuker(&V);
				break;

      case kFitch:
				strcpy(V.seqid, V.s);
				getline(&V);
				readFitch(&V);
				break;
			
			case kOlsen:
				readOlsen(&V);
				break;
				
      case kGCG:
				do {
					gotuw = (strstr(V.s,"..") != NULL);
					if (gotuw) readUWGCG(&V);
					else adddoc(V.s, &V);
					getline(&V);
				} while (!(Feof(V.f) || V.allDone));
				break;
      }
  	} 
		
	else { /* format checking */
    /* check for ".." of uwgcg, since it can masquerade as any
            other format */
		/* 6Mar90 -- added G.Olsen multisequence format */
    i = 0;
    do {
      /*! check most of file for ".." */
			/* ECOLAC GenBank document header is 300 lines !*/
			
      getline(&V); i++;
			gotolsen = (strstr(V.s, "identity:   Data:") != NULL);
			gotuw = (strstr(V.s, "Check:") != NULL);
      if (gotuw) gotuw = (strstr(V.s,"..") != NULL);
  		else adddoc(V.s, &V);  /* must save text here for GCG */
   	  V.done = (gotolsen || gotuw || Feof(V.f) || i > 800);
      if (gotuw && (*V.s == ';')) {
				/* fix for ToIG of a GCG sequence */
				gotuw = false;
				V.done = true;    
      	}
    } while (!V.done);

		
		if (gotolsen) readOlsen(&V);
		else if (gotuw) readUWGCG(&V);
    else {
			*V.doclen = 0;  /*drop for non-uw */
      Frewind(V.f);
      do {
				getline(&V);
				l = strlen(V.s);
				while (l > 0 && V.s[--l] == ' ') ;
      } while (!((l != 0) | Feof(V.f)));
      if (l == 0) V.err = eNoData;
    }

    if (!(V.err != 0 || V.allDone)) {  
      if (*V.s == ';') {
				if (strstr(V.s,"Strider") != NULL)
					readStrider(&V);
				else
					readIG(&V);
				} 
			else if (strstr(V.s,"LOCUS") == V.s)
					readGenBank(&V);
			else if (*V.s == '>') {
					if (V.s[3] != ';') readPearson(&V);
					else readNBRF(&V);
					}
			else if (strstr(V.s,"ID   ") == V.s)
					readEMBL(&V);
			else if (*V.s == '(')
					readZuker(&V);
			else {
				strcpy(V.seqid, V.s);
				getline(&V); /*test for fitch format*/
				i = 1;
				do {
					isfitch = (((i - 1) & 3) == 0 && V.s[i - 1] == ' ' ||
							 ((i - 1) & 3) != 0 && V.s[i - 1] != ' ');
					i++;
				} while (i < strlen(V.s) && isfitch);
				if (isfitch) readFitch(&V);
				else readUnknown(&V);
				}
			}
		}

  if (V.choice == 0) *V.seqlen = V.ninfo;
  V.seq[*V.seqlen] = 0; /* stick a string terminator on it */
  V.doc[*V.doclen] = 0; 	/* " */
	Fclose(V.f);  
  return V.err;
}  /*readSeq*/




Handle readFile(XCmdPtr pXCmd, char *fname)
{
	Point				gLocation = {0x40,0x40};
	SFReply			reply;			/* used in all SF samples */
	SFTypeList	typeList;		/* typelist for all SF samples */
	Handle			hdata, hvref;
	long				hlen, vref;
	short				err, fref;
	
	if (*fname != 0) {
	/* ! need to save/read reply.vRefNum b/n calls */
	  hvref = GetGlobal(pXCmd,"\pvrefnum");
		if (hvref != NULL) {
			stringtonum(*hvref,&vref);
			DisposHandle(hvref);
			}
		else vref = 0;
		if (fsopen( fname, vref, &fref) != 0) 
		  return NULL;
		}
	else {
		typeList[0] = 'TEXT';
		SFGetFile(gLocation, "\pooh", nil, 1, &typeList, nil, &reply);	
		if (reply.good) { 
			strncat( fname, &(reply.fName[1]), reply.fName[0]);
			if (FSOpen( reply.fName, reply.vRefNum, &fref) != 0) 
				return NULL;
			hvref = NewHandle(50);
			numtostring(reply.vRefNum, *hvref);
			SetGlobal( pXCmd, "\pvrefnum", hvref);
			DisposHandle(hvref);
			}
		else 
			return NULL;
		}
		
	err = GetEOF( fref, &hlen);
	hdata = NewHandle(hlen);
	HLock( hdata);
	err = FSRead( fref, &hlen, *hdata);
	err = FSClose( fref);
	HUnlock( hdata);
	return hdata;
}


void CallReadSeq(XCmdPtr pXCmd)
{
	short			np, nseq, format, err = 0;
	char			seqid[256], fname[256], **pname, **pdoc, **c;
  long 			seqlen, dlen, iseq, doclen;
	Handle		hseq, hdata, hdoc;
	Boolean		passname, passdoc, passdata, cdname, cddoc;
	Str255		str;
	
	np = pXCmd->paramCount;

	c = (pXCmd->params[0]);
	if (np > 0 & c != NULL & **c != 0) stringtonum( *c, &iseq);
	else iseq = 1;
		
	passname = false; *fname = 0;
	passdoc	= false;	 
	passdata = false; 
	
	if (np > 2) {
		ZeroToPas( pXCmd, *(pXCmd->params[1]), str);
		cdname = StrToBool( pXCmd, str);
		passname= (pXCmd->params[2] != NULL);
		pname	= (pXCmd->params[2]);
		/* !!?? do we never want to use field name to read file */
		ZeroToPas( pXCmd, *pname, str);
		hdoc = GetFieldByName( pXCmd, cdname, str);
		strcpy( fname, *hdoc);
		DisposHandle( hdoc);
		}
		
	if (np > 4) {
		ZeroToPas( pXCmd, *(pXCmd->params[3]), str);
		cddoc = StrToBool( pXCmd, str);
		passdoc = (pXCmd->params[4] != NULL);
		pdoc	= (pXCmd->params[4]);
		}

	if (np > 5) {
		hdata = (pXCmd->params[5]);
		passdata = (hdata != NULL & **hdata != 0);
		}
	if (!passdata) hdata = readFile(pXCmd, fname);
	if (hdata == NULL) { /* no read */
		/* pXCmd->returnValue = hdata; */ /* does hc set this? */
		return;
		}
		
	dlen		= GetHandleSize(hdata);
	seqlen 	= min( MAXSEQ, dlen);
	hseq 		=	NewHandle(seqlen);
	doclen	= min( MAXDOC, dlen);
	hdoc 		= NewHandle(doclen);
	format 	= 0;   /*!*/

	HLock( hdata);
	HLock( hseq); 
	HLock( hdoc);
	err = readSeq( iseq, *hdata, dlen, *hseq, &seqlen, &nseq, 
									&format, *hdoc, &doclen, seqid);
	HUnlock( hdoc);
	HUnlock( hseq);
	HUnlock( hdata);
	if (!passdata) DisposHandle( hdata);
	SetHandleSize( hseq, seqlen+1);
	SetHandleSize( hdoc, doclen+1);
	
	if (passdoc) { 
		ZeroToPas( pXCmd, *pdoc, str);
		SetFieldByName(pXCmd, cddoc, str, hdoc);
		}
	DisposHandle(hdoc);

	if (passname) {
		hdoc = NewHandle(strlen(fname)+1);
		strcpy( *hdoc, fname);
		ZeroToPas( pXCmd, *pname, str);
		SetFieldByName(pXCmd, cdname, str, hdoc);
		DisposHandle(hdoc);
		}

	pXCmd->returnValue = hseq;  
}
