/*writeseqXCMD.c
 *
 * Copyright 1990 by d.g.gilbert.
 * dogStar Software && Indiana University Biology Dept.
 * email: gilbertd@iubio.bio.indiana.edu
 *
 * from readseq.c
 *
 * Language: MPW-C 3.0 
 *

	 HCard syntax is:
--
--	writeseq  
--		seq,			: container w/ seq
--		[#	]			: format to convert to, GCG is default 
--		[doc]			: container w/ document, if any
--	the result returns formatted seq

*/

/*-------- Makefile

C -b writeseqXCMD.c -mbg off
Link -w -rt XCMD=10020 
	-m ENTRYPOINT 
	-sg writeseq 
	writeseqXCMD.c.o 
	"{Libraries}HyperXLib.o" 
	"{Libraries}Interface.o" 
	"{CLibraries}StdCLib.o" 
	"{Clibraries}"Cinterface.o 
	"{CLibraries}CRuntime.o" 
	-o "HyperGCG" 
#	-o "writeseq.xcmd" -x writeseq.xref	-map > writeseq.map
#-------------------------*/


#include <CType.h>
#include <String.h>
#include <Types.h>
#include <StdLib.h>
#include <HyperXCmd.h>

#include <Memory.h>
#include <OSUtils.h>
#include <ToolUtils.h>
#include <Packages.h>



/* XCMD entry -- MUST BE FIRST CODE IN SEG */

void 	CallWriteSeq(XCmdPtr pXCmd);

pascal void EntryPoint(XCmdPtr pXCmd)
{
	CallWriteSeq(pXCmd);
	return;
}


#define Local      static    /* local functions */

#define	NNB				18
#define aminos		"ABCDEFGHIKLMNPQRSTVWXYZ*"
#define iubbase	 	"ACGTUMRWSYKVHDBXN." 
#define igbase	 	"ACGTUJRLMYKNNNNNN?"
#define primenuc	"ACGTU"
#define protonly	"EFIPQZ"

typedef	char	boolean;
#define NEWLINE       13  
#define	false	0
#define	true	1
#define min(a,b)  		(a<b)?a:b
#define max(a,b)  		(a>b)?a:b
#define _toupper(c)		((c)-'a'+'A')
#define _tolower(c)		((c)-'A'+'a')
#define	Isalpha(c)		((c >= 'A' & c <= 'Z') \
										|  (c >= 'a' & c <= 'z'))
#define	Islower(c)		((c >= 'a') && (c <= 'z'))


#define MAXSEQ			32000		/* HCard field max chars */


/* formats known to readSeq */
#define kIG             1
#define kGenBank        2
#define kNBRF           3
#define kEMBL           4
#define kGCG            5
#define kStrider        6
#define kFitch          7
#define kPearson        8
#define kZuker          9
#define kOlsen          10
#define kUnknown        11
#define kMinFormat      1
#define kMaxFormat      11



int Toupper(char c)
{
	return( (c>='a')&&(c<='z') ? (c-('a'-'A')) : c );
}


void upcase(char *s)
{
	int i,l;
	l = strlen(s);
	for (i=0;i<l;i++) s[i] = Toupper(s[i]);
}
		    
				
#define allsymbols  "_.-*?<>{}[]()!@#$%^&=+;:'/|`~\"\\" 

#define isseqchar(c) (Isalpha(c) || strchr(allsymbols, c))


short GCGchecksum( char *seq, long seqlen)
{
	long	check = 0, count = 0, i;

	for (i = 0; i < seqlen; i++) {
		count++;
		check += count * Toupper(seq[i]);
		if (count == 57) count = 0;
		}
	return (check % 10000);
}	


boolean isNucleic( char *seq, long seqlen)
{ /* return sequence kind: Nucleic (true) or Amino (false) */
	char	c;
	short	i, na, aa, maxtest;
	
	maxtest = min(200, seqlen);
	for (i = 0, na = 0, aa = 0; i < maxtest; i++) {
		c = seq[i];
		if (Islower(c)) c = _toupper(c);
		if (strchr(protonly, c)) return(false);
		else if (strchr(primenuc,c)) na++;
		else if (strchr(aminos,c)) aa++;
		}
	return (na > aa);
} /* isNucleic */


Local void mapbase( char *seq, long seqlen,
									  char *frombase, char *tobase)
{ /* translate ambiguity codes b/n ig and iub/gcg */
	long	i;
	char	c;
	boolean	islow;
	short	j;
	
	if (isNucleic(seq, seqlen))
	 for (i = 0; i < seqlen; i++) {
		c = seq[i];
		islow = Islower( c);
		if (islow) c = _toupper(c); 
		for (j = 0; j < NNB; j++) if (c == frombase[j]) {
			c = tobase[j];
			if (islow) seq[i] = _tolower(c);
			else seq[i] = c;
			break;
			}
		}
}



void writeSeq(char *seq, long seqlen, 
							char *doc, char *seqout, 
							/* ?? add idword == filename */ 
							int inform, int outform)
/* put formatted sequence to seqout */
/* !! can't use sprintf() in XCMD -- uses Global data */
{ 
	short	spacer = 0, width	= 50, tab = 0;
	boolean numline = false, space1 = true;
	char	idword[31], endstr[10];
	char  s[100], seqid[100], snum[30], stime[50];
	char	*c, *b;
	long	i, j, k, l1, ibase;
	unsigned long today;
	 
	
	if (doc == NULL) strcpy(seqid, "Untitled");
	else {
	  while (*doc == ' ' & *doc == '\n') doc++;
		for (	i=0; 
					i<80 & doc[i] != 0 & doc[i] != '\n';
					i++)
						seqid[i] = doc[i];
		seqid[i] = 0;
		}
	for ( j=0; 
				j<30 & seqid[j] != ' ' & seqid[j] != 0; 
				j++)
					idword[j] = seqid[j];
	idword[j] = 0;
	
	strcpy( endstr,"");
	strcpy( seqout,"");
	numtostring(seqlen, snum);
	
	strcpy(stime, "  ");
	GetDateTime( &today);
	iudatestring( today, abbrevDate, s);
	strcat(stime, s);
	strcat(stime, "  ");
	iutimestring( today, false, s);
	strcat(stime, s);
	
	switch (outform) {
	
		case kOlsen:	/* Olsen seq. editor only takes plain nucs */
		case kUnknown:  	/* no header, just sequence */
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			strcpy(endstr,"\n"); /* end w/ extra blank ibase */
			spacer = 11; /* ?? want this for our card data */
			space1 = false; /* no 1st space for us */
			break;  
			
		case kGenBank:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			strcat(seqout, "LOCUS       ");
			strcat(seqout, idword);
			strcat(seqout, "       ");
			strcat(seqout, snum);
			strcat(seqout, " bp ");
			strcat(seqout, stime);
			strcat(seqout, "\n");
			
			strcat(seqout, "DEFINITION  ");
			strcat(seqout, seqid);
			strcat(seqout, "\n");
	 /* printf("ACCESSION   %s\n", accnum); */
			
			/* put all of doc as COMMENT */
			b = doc;
			if (doc != NULL) {
				strcat(seqout, "COMMENT \n");
				while (*b != 0) {
					while (*b == '\n') b++;
					while (*b == ' ') b++;
					for (c = b, j = 0; j<68 & *c!=0 & *c!='\n'; c++, j++);
					strcat(seqout, "   ");
					strncat( seqout, b, j);
					strcat(seqout,"\n");
					b = c;
					}
				}
				
			strcat(seqout, "ORIGIN      \n");
			spacer = 11;
			numline	= true;
			strcpy(endstr, "\n//");
			break;
			
		case kNBRF:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			if (isNucleic(seq, seqlen)) 
				strcat(seqout,">DL;");
			else 
				strcat(seqout,">P1;");
			strcat(seqout,idword);
			strcat(seqout, "\n");
			strcat(seqout, seqid);
			strcat(seqout, ", ");
			strcat(seqout, snum);
			strcat(seqout, " bases.\n\n");
			spacer = 11;
			numline	= true;
			strcpy(endstr,"*");
			break;
			
		case kEMBL:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			strcat(seqout,"ID   ");
			strcat(seqout, idword);
			strcat(seqout,"\n");
	/*	printf("AC   %s\n", accnum); */
			strcat(seqout,"DE   ");
			strcat(seqout, seqid);
			strcat(seqout,"\n");
			strcat(seqout,"SQ   		       ");
			strcat(seqout, snum);
			strcat(seqout," BP\n");
			break;
			
		case kGCG:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			/* drop any existing terminators in doc */
			while (NULL != (b = strstr(doc,".."))) *b = ',';
			/* put all doc at top ... need to check for newlines*/
			b = doc;
			if (doc != NULL) while (*b != 0) {
				while (*b == '\n') b++;
				for (c = b, j = 0; j<80 & *c!=0 & *c!='\n'; c++, j++);
				strncpy( s, b, j); s[j] = 0;
				if (NULL == strstr(s, "Check:")) {
					strcat( seqout, s);
					strcat( seqout,"\n");
					}
				b = c;
				}
	 /* printf("ACCESSION   %s\n", accnum); */
	/*	printf("    %s  Length: %d  (today)  Check: %d  ..\n",
      				idword, seqlen, GCGchecksum(seq, seqlen)); */
			strcat(seqout, "    ");
			strcat(seqout, idword);
			strcat(seqout, "  Length: ");
			strcat(seqout, snum);
			strcat(seqout, stime);
			strcat(seqout, "  Check: ");
			numtostring(GCGchecksum(seq, seqlen), snum);
			strcat(seqout, snum);
			strcat(seqout, "  ..\n");
			
			spacer = 11;
			numline	= true;
			break;
			
		case kStrider: /* ?? map ?*/
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			strcat(seqout,"; ### from DNA Strider ;-)\n");
			strcat(seqout,"; DNA sequence  ");
			strcat(seqout, seqid);
			strcat(seqout, ", ");
			strcat(seqout, snum);
			strcat(seqout," b.p.\n");
			strcat(seqout,";\n");
			strcpy(endstr, "\n//");
			break;
			
		case kFitch:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			strcat(seqout, seqid);
			strcat(seqout,", ");
			strcat(seqout,snum);
			strcat(seqout," bases.\n");
			spacer = 4;
			width	= 60;
			break;
			
		case kIG:
			if (inform != kIG) mapbase(seq, seqlen, iubbase, igbase);
			strcat(seqout,"; ");
			strcat(seqout, seqid);
			strcat(seqout,", ");
			strcat(seqout,snum);
			strcat(seqout," bases.\n");
			/* put all doc at top...*/
			b = doc;
			if (doc != NULL) while (*b != 0) {
				while (*b == '\n') b++;
				for (c = b, j = 0; j<80 & *c!=0 & *c!='\n'; c++, j++);
				strcat(seqout,";"); /* comment line */
				strncat( seqout, b, j);
				strcat(seqout,"\n");
				b = c;
				}
			strcat(seqout, idword);
			strcat(seqout, "\n");
			strcpy(endstr,"1"); /* == linear dna */
			break;
			
  	default :
		case kZuker: /* don't attempt Zuker's ftn format */
		case kPearson:
			if (inform == kIG) mapbase(seq, seqlen, igbase, iubbase);
			/*printf(">%s, %d bases\n", seqid, seqlen);*/
			strcat(seqout,">");
			strcat(seqout, seqid);
			strcat(seqout,", ");
			strcat(seqout,snum);
			strcat(seqout," bases.\n");
			break;
		}
		
	
	width = min(width,100);
	b = seq; c = s; l1 = 0; ibase = 1; 
	while ( *b != 0 ) {
		if (numline && (l1 == 0)) {
			numtostring(ibase, snum);
			k = 8 - strlen(snum);
			for (j=0; j<k; j++) strcat(seqout," ");
			strcat(seqout, snum);
			strcat(seqout, " ");
			}
				
		if ((spacer != 0) && (space1 | l1 > 0) 
					&& ((l1+1) % (spacer-1) == 1)) 
				*c++ = ' '; 
		*c++ = *b++; 
		ibase++;
		l1++;
		if (l1 == width || *b == 0) {
			for (j=0; j<tab; j++) strcat(seqout," ");  
			*c = 0;   
			strcat(seqout,s);
			if (*b == 0) strcat(seqout,endstr);
			strcat(seqout,"\n");
			c = s; l1 = 0;
			}
		}
}  /*writeSeq*/




void 	CallWriteSeq(XCmdPtr pXCmd)
{
	int			seqlen, kform, np, i, k;
	char		**c, *pdoc, *b, *seq;
	Handle	hdoc, hseq, hout;
	
	
	np = pXCmd->paramCount;
	
	hseq = (pXCmd->params[0]);
	if (np > 0 & hseq != NULL & **hseq != 0) ;
	else { /* fail */
		/* pXCmd->returnValue = hseq; *//* ?NULL to HC */  
		return;
		}
		
	/* ! filter out junk from seq */
	HandToHand( &hseq); /* dup so we can mangle it */
	HLock(hseq);
	seq = *hseq;
	for (i = 0, b = seq; seq[i] != 0; i++)
		if isseqchar(seq[i]) *b++ = seq[i];
	*b = 0;  
	seqlen = (long) b - seq;
	HUnlock(hseq);
	SetHandleSize(hseq, seqlen+1);
	
	c = (pXCmd->params[1]);
	if (np > 1 & c != NULL & **c != 0) stringtonum( *c, &kform);
	else kform = kGCG; /* default */

	hdoc = (pXCmd->params[2]);
	if (np > 2 & hdoc != NULL & **hdoc != 0) { 
		HLock(hdoc);
		pdoc = *hdoc;
		}
	else pdoc = NULL;
	
	k = GetHandleSize(hseq) + GetHandleSize(hdoc);
	hout = NewHandle( 2*k); /* ?? expect k <= 32000 */
	
	HLock(hseq);
	HLock(hout);
	writeSeq( *hseq, seqlen, pdoc, *hout,
						kUnknown, /* always passing unformatted data ? */
						kform);
	HUnlock(hseq);
	if (pdoc != NULL) HUnlock(hdoc);
	HUnlock(hout);
	DisposHandle(hseq); /* drop our copy */
	SetHandleSize( hout, 1+strlen(*hout));
	
	pXCmd->returnValue = hout;  
	return;
}

